var fs = require('fs');
var path = require('path');
var glob = require('glob');
var tpl = require('./template');
var htmlparser = require('htmlparser2');

const SA_TAG = '_elong_sa_tag';
var LF = {
    data: '\n',
    type: 'text'
};
var { DomUtils } = htmlparser;
var { find } = DomUtils;
var getDom = code => htmlparser.parseDOM(code, { xmlMode: true });
var getHtml = dom => DomUtils.getOuterHTML(dom, { xmlMode: true });
var removeLF = code => code.split('\n').filter(l => l.trim()).join('\n');
var findTag = (test, root, all = 1) => DomUtils[all ? 'findAll' : 'findOne'](test, root);
var removeComment = dom => find(n => n.type === 'comment', dom, true).forEach(ele => DomUtils.removeElement(ele));

var Compiler = {
    build(dir = process.cwd(), tap) {
        glob(path.resolve(dir, "**/*.wxml"), (er, files) => {
            Compiler.praseFiles(files, tap, _ => console.log('total ', files.length, ' all done ...'));
        });
    },
    praseFiles(files, tap, cb) {
        var total = files.length;
        files.forEach(key => {
            fs.readFile(key, 'utf8', (err, data) => {
                var code;
                try {
                    data.includes('_elong_sa_touchStart') || (code = Compiler.addWrap(data));
                    code = tap ? Compiler.parseTap(code || data) : code ? getHtml(code) : code;
                    if (code) {
                        fs.writeFile(key, removeLF(code), (err, data) => {
                            console.log(key, ' build ok ');
                            --total || cb();
                        });
                    } else {
                        --total || cb();
                    }
                } catch (e) {
                    console.log(e, ' key = ', key);
                }
            });
        });
    },
    addWrap(code) {
        var dom = getDom(`<view>${tpl(code)}</view>`);
        var wrap = findTag(n => n.attribs.bindtouchstart === '_elong_sa_touchStart', dom, 0);
        wrap.children.forEach(ele => {
            if (ele.name === 'import') {
                DomUtils.removeElement(ele);
                DomUtils.prepend(wrap, ele);
                DomUtils.prepend(wrap, LF);
            }
        });
        removeComment(dom);
        return dom[0].children;
    },
    parseTap(code) {
        var dom = Object.prototype.toString.call(code) === '[object String]' ? getDom(code) : code;
        findTag(n => n.attribs.bindtap || n.attribs.catchtap, dom).forEach(Compiler.addId4Node);
        return getHtml(dom);
    },
    addId4Node(node, i) {
        if (node.attribs.id) return;
        node.attribs.id = (node.attribs.bindtap || node.attribs.catchtap) + "_" + i;
        node.attribs['data-' + SA_TAG] = node.attribs.bindtap ? 'bindtap' : 'catchtap';
        var parent = node.parent;
        if (!parent) return;
        var iChain = [];
        do {
            if (parent.attribs['wx:for']) {
                let def = parent.attribs['wx:for-index'] || 'index',
                    toIndex = "index_" + i + "_" + iChain.length;
                Compiler.modifyIndexIdentifier(parent, def, toIndex);//修改index标识符
                iChain.push(parent.attribs['wx:for-index'] = toIndex);
            }
        } while ((parent = parent.parent));
        if (iChain.length) {
            node.attribs.id += "_{{" + iChain.join(" + '_' + ") + "}}";
        }
    },
    modifyIndexIdentifier(node, raw, cooked, filter = []) {
        var rp = str => Compiler.replace(str, raw, cooked, filter);
        switch (node.type) {
            case 'text':
                node.data = rp(node.data);
                break;
            case 'tag':
                Object.keys(node.attribs).forEach(key => node.attribs[key] = rp(node.attribs[key]));
                node.children.forEach(obj => {
                    var hasPush = 0;
                    if (obj.type === 'tag' && obj.attribs['wx:for'] && ++hasPush) {
                        filter.push(obj.attribs['wx:for-index'] || 'index');//作用域入栈
                    }
                    Compiler.modifyIndexIdentifier(obj, raw, cooked, filter);
                    hasPush && filter.pop();//作用域出栈
                });
                break;
        }
    },
    replace(str, raw, cooked, filter = []) {
        let reg = new RegExp('({{(?:\s|.)*)(\\b' + raw + '\\b)((?:\s|.)*}})');
        if (reg.test(str) && !filter.includes(RegExp.$2)) {
            while (reg.test(str)) str = str.replace(reg, `${RegExp.$1}${cooked}${RegExp.$3}`);
        }
        return str;
    }
}
module.exports = Compiler.build;